/*
 * 161plus.c by xt (xt@codeslum.org)   (c) late september - mid october, 2011
 * http://xt.codeslum.org/161plus/
 *
 * this is a modified version of: onesixtyone version 0.3.2 (c) 2002,2003 Solar Eclipse (solareclipse@phreedom.org)
 *
 * improvements/changes:
 *
 *    *) extreme speed increase: instead of 1 select() per packet, the amount is now user-configurable (-p); default is 250.
 *       per my last benchmark we can now scan 1 million IPs, reliably, in just 56 seconds with -p set to 1000 on my 100mbit pipe.
 *    *) support for SNMP v1 GET and SNMP v2c GETBULK (-v 1|2). Default is 1.
 *    *) all snmp error message output is prefixed with "(error)" so that you can remove useless snmp responses from your logs
 *       with grep, such as grep -v "(error)" logfile.
 *    *) includes cidr generation script; use ./cidr 24.217.0.0/16 and it will output all IPs in that CIDR range.
 *    *) supports reading variable number of IPs from IP file (sort of; (st_size / 10) + 1000) should handle most IP files, with a lot of room
 *       to spare, unless it's full of a bunch of 1 byte dotted-decimal segments instead of an amount 1 - 4 (but adding 1000 helps fix this).
 *       this is done because i want to read through the IP list only once. most machines should have enough memory nowadays. if yours
 *       doesn't just split your file with split(1).
 *    *) randomizes ip scan order when -r option is specified, to help avoid NIDS going nuts.
 *    *) error output, debug output, and others are cleaned up and perror()/strerror() is used to make sense of it.
 *    *) other functional, cosmetic (both output and code) changes like signal trapping and better error handling
 *   
 *   example:
 *     FIRST, generate IPs to scan:
 *       ./cidr 206.0.0.0/14 > 206.0-14
 *         to generate all IPs in 206.0.0.0/14 and put in file 206.0-14
 *     SECOND, do the scan!
 *       ./161plus -i 206.0-14 -o 206.0-14.output -r -p 1000 public
 *         to run the scanner on all the ips you just generated with community "public", in random order, at a very fast rate and write
 *         out to 206.0-14.output.
 *     THIRD, profit?
 *       do whatever you want with the results! >:-)
 *
 *     check the #defines and make sure everything is to your liking. tested and known to compile on Linux, FreeBSD and Mac OS X.
 *     should also compile on NetBSD, OpenBSD, and Solaris.
 *
 *     shouts to wx, zorch, subz, asd, ghost, voronika, substance, doomed, james, booyah, cidr, merlin, hockey, cl0wn, bahb,
 *               osb, shahid, presonic, morgan, wd, sean, my girlfriend, my son, the baby on the way and any other people i forgot.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <math.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define MAX_COMMUNITIES 128		/* Maximum number of communities; change for use with large dictionary files. */
#define MAX_COMMUNITY_SIZE 32		/* Maximum length of community string. */
#define MAX_ERROR 10			/* Maximum number of errors with sending of packets, */
#define USLEEP_TIME 30			/* Value for usleep() */
#define EXTRASIZE 1000			/* Extra number of IPs to allocate memory for (to compensate for shitty read-once IP read function). */

/* ** */

#ifdef SOLARIS	/* Solaris needs a u_int32_t type .. hope this works. Thanks yuckf00! */
	typedef unsigned long u_int32_t;
#endif

char* snmp_errors[] = {
	"NO ERROR",				/* 0 */
	"TOO BIG",				/* 1 */
	"NO SUCH NAME",			/* 2 */
	"BAD VALUE",			/* 3 */
	"READ ONLY",			/* 4 */
	"GENERIC ERROR",		/* 5 */
	"NO ACCESS",			/* 6 */
	"WRONG TYPE",			/* 7 */
	"WRONG LENGTH",			/* 8 */
	"WRONG ENCODING",		/* 9 */
	"WRONG VALUE",			/* 10 */
	"NO CREATION",			/* 11 */
	"INCONSISTENT VALUE",	/* 12 */
	"RESOURCE UNAVAILABLE",	/* 13 */
	"COMMIT FAILED",		/* 14 */
	"UNDO FAILED",			/* 15 */
	"AUTHORIZATION ERROR",	/* 16 */
	"NOT WRITABLE",			/* 17 */
	"INCONSISTENT NAME",	/* 18 */
};

/* Program options */
struct {
	int debug;
	int log;
	int quiet;
	long wait;	/* time to wait for select() */
	int packets;	/* # of packets before calling routine to check select() */
	int version;	/* 1 or 2 (2c) */
	int usl;	/* usleep() */
	int randomize;

	FILE* log_fd;
} o;

char *community[MAX_COMMUNITIES] = { "public", NULL };
int community_count = 1;

u_int32_t *hosts = NULL;
unsigned long host_count = 0UL;
unsigned long host_size = 0UL;
unsigned long host_max = 0UL;

#ifndef INADDR_NONE			/* Solaris is broken */
	#define INADDR_NONE -1
#endif

void sig_catch(int);
void array_shuffle(u_int32_t *array, size_t n);

/* ** */

void usage()
{
	printf("usage: 161plus [options] [host] [community]\n");
	printf("\noptions:\n");
	printf("\t-c <communityfile> file with community names\n");
	printf("\t-i <inputfile>     file with hosts (for mass-scanning)\n");
	printf("\t-o <outputfile>    output log\n");
	printf("\t-d                 debug mode; use twice for more information\n\n");
	printf("\t-w n               select() timer, default 30; keep in range 10-100.\n");
	printf("\t-p n               # of packets before select/recv() loop (dflt 250)\n");
	printf("\t-u n               usleep(n) value after each scan loop (-p); default 10)\n");
	printf("\t-v 1|2             choose snmp v1 or v2c; v2c checks for BULK get, 1 for GET\n");
	printf("\t-r                 enable ip scan order randomization\n");
	printf("\t-q                 quiet mode, do not print log to stdout, use with -o\n");
	printf("\ntip: set -p to lower value (~50) when scanning lists < 75,000 IPs.\n");
}

void read_communities(char* filename)
{
	FILE* fd;
	int i, c;
	char ch;

	if (o.debug > 0) fprintf(stderr,"Using community file %s\n", filename);
		
	if ((fd = fopen(filename, "r")) == 0) {
		fprintf(stderr,"Error opening community file \'%s\' (%s)\n", filename, strerror(errno));
		exit(EXIT_FAILURE);
	}

	i = 0; c = 0;
	community[i] = (char*)malloc(MAX_COMMUNITY_SIZE);
	while ((ch = fgetc(fd)) != EOF) {
		if (ch == '\n') {
			community[i][c] = '\0';
			if (c > 0) {
				i++; c = 0;
				if (i >= (MAX_COMMUNITIES - 1))
				{
					fprintf(stderr, "Warning: tToo many communities (%d); not reading anymore.\n", i);
					fclose(fd);
					return;
				}
				else
					community[i] = (char*)malloc(MAX_COMMUNITY_SIZE);
			}
		} else if (ch == '\r') {
			c++;
			continue;
		} else {
			community[i][c++] = ch;
		}
		if (c > MAX_COMMUNITY_SIZE-1) {
			fprintf(stderr,"Error: Community string too long. Please make sure all communities are less than %d bytes.\n", MAX_COMMUNITY_SIZE);
			exit(EXIT_FAILURE);
		}
	}
			
	community_count = i;
	fclose(fd);
}

void read_hosts(char* filename)
{
	FILE* fd;
	char buf[24];
	char ch;
	int c;
	struct stat ipstat;

	memset(&ipstat, 0, sizeof(struct stat));

	if (strcmp(filename, "-") == 0) {
		fprintf(stderr, "stdin input is no longer supported; please use a filename.\n");
		exit(EXIT_SUCCESS);
	}
	else {
		unsigned int rounded = 0U;
		if (o.debug > 0) fprintf(stderr,"Reading hosts from input file \'%s\'\n", filename);
		if (stat(filename, &ipstat) == -1)
		{
			perror("stat(file with ips)");
			exit(EXIT_FAILURE);
		}

		rounded = roundf(ipstat.st_size / 10); // This is extremely crude, but should work.
		hosts = (u_int32_t *) malloc(sizeof(u_int32_t) * (rounded + EXTRASIZE)); // An extra 1000 for good measure.
		host_size = sizeof(u_int32_t) * (rounded + EXTRASIZE);
		host_max = rounded + EXTRASIZE;
		if (o.debug) fprintf(stderr, "Number of host entries allocated: %lu\n", host_max);
		if ((fd = fopen(filename, "r")) == 0) {
			fprintf(stderr,"Error: Cannot open input file \'%s\' (%s)\n", filename, strerror(errno));
			free(hosts);
			exit(EXIT_FAILURE);
		}
	}

	host_count = 0UL;
	c = 0; ch = 0;

	while ((ch = fgetc(fd)) != EOF) {
		if (ch == '\n') {
			buf[c] = '\0';
			if (c > 0) {			/* skip blank lines */
				if ((host_count >= (host_max-1)))
				{
					fprintf(stderr, "Warning: Too many IPs; loaded %lu IPs, not loading anymore.\n", host_count);
					return;
				}
				else
				{
					if ((hosts[host_count++] = inet_addr((const char*)&buf)) == INADDR_NONE) {
						fprintf(stderr,"Error: Malformed IP address: %s\n", buf);
						exit(EXIT_FAILURE);
					}

					c = 0;
				}
			}
		} else if (ch == '\r') {
			c++;
			continue;
		} else {
			if (!isdigit(ch) && ch != '.')
			{
				fprintf(stderr, "Error: Invalid character in IP file (%d), exiting. Please fix file.\n", ch);
				exit(EXIT_FAILURE);
			}
			else
				buf[c++] = ch;
		}

		if (c > (sizeof(buf) - 1)) {
			fprintf(stderr, "Error: IP address exceeds buffer size; IP file is in improper format.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (o.randomize)
	{
		array_shuffle(hosts, host_count);
		if (o.debug) { fprintf(stderr, "Shuffled scan IPs.\n"); }
	}

	fclose(fd);

	if (o.debug > 0) fprintf(stderr,"%lu hosts read from file\n", host_count);
}

void init_options(int argc, char *argv[])
{
	char community_filename[250];
	char input_filename[250];
	char log_filename[255];
	int input_file;
	int community_file;
	int do_exit = 0;

	int arg, i;
	
	o.debug = 0;
	o.log = 0;
	o.quiet = 0;
	o.wait = 10;
	o.packets = 250;
	o.version = 1;
	o.usl = 10;
	o.randomize = 0;
	input_file = 0;
	community_file = 0;

	o.log_fd = NULL;

	while ((arg = getopt(argc, argv, "c:di:o:w:p:v:u:qr")) != EOF) {
		switch (arg) {
			case 'c' :	community_file = 1;
						strncpy(community_filename, optarg, sizeof(community_filename));
						break;
			case 'd' :	o.debug++;
						break;
			case 'i' :	input_file = 1;
						strncpy(input_filename, optarg, sizeof(input_filename));
						break;
			case 'o' :	o.log = 1;
						strncpy(log_filename, optarg, sizeof(log_filename));
						break;
			case 'w' :  o.wait = atol(optarg);	/* convert to nanoseconds */
						break;
			case 'q' :	o.quiet = 1;
						break;
			case 'p' :
					o.packets = atoi(optarg);
						break;
			case 'u' :
					o.usl = atoi(optarg);
						break;
			case 'v' :
					o.version = atoi(optarg);
						break;
			case 'r' :
					o.randomize = 1;
						break;
			case '?' :  usage();
						exit(EXIT_SUCCESS);
						break;
		}
	}

	if (o.usl <= 0)
	{
		fprintf(stderr, "Error: invalid -u argument %d. Must be greater than 0 (default 10).\n", o.usl);
		do_exit = 1;
	}
	if (o.version != 1 && o.version != 2)
	{
		fprintf(stderr, "Error: invalid -v argument %d. Must specify either 1 or 2.\n", o.version);
		do_exit = 1;
	}

	if (o.packets <= 0 || o.packets > 10000)
	{
		fprintf(stderr, "Error: invalid -p argument %d. Packets must be > 0 && < 10000.\n", o.packets);
		do_exit = 1;
	}

	if (do_exit) { exit(EXIT_FAILURE); }

	if (o.debug) {
		if (o.debug > 0) fprintf(stderr,"Debug level %d\n", o.debug);
	}

	if (!input_file) {
		if (optind >= argc) { 
			usage();
			exit(EXIT_SUCCESS);
		}

		hosts = (u_int32_t *) malloc(sizeof(u_int32_t) * 1);
		if ((hosts[0] = inet_addr((const char*)argv[optind++])) == INADDR_NONE) {
			fprintf(stderr, "Error: Malformed IP address: %s\n", argv[optind-1]);
			free(hosts);
			exit(EXIT_FAILURE);
		}
		host_count = 1;
		host_size = sizeof(u_int32_t) * 4;
		host_max = 1;
		if (o.debug > 0) fprintf(stderr,"IP read from the command line: %s\n", argv[optind-1]);
	}
	else {
		read_hosts((char*)&input_filename);
	}

	if (community_file) {
		read_communities((char*)&community_filename);
	}

	if (optind < argc) {
		if (community_file) {
			usage();
			exit(EXIT_SUCCESS);
		}
		community[0] = argv[optind++];
		community_count = 1;
		if (o.debug > 0) fprintf(stderr,"Community read from command line: %s\n", community[0]);;
	}

	if (optind < argc) {
		usage();
		exit(EXIT_SUCCESS);
	}
	
	if (o.log) {
		if ((o.log_fd = fopen(log_filename, "a")) == 0) {
			fprintf(stderr,"Error opening log file \'%s\' (%s)\n", log_filename, strerror(errno));
			exit(EXIT_FAILURE);
		}
		printf("Logging to: %s\n", log_filename);
	} else if (o.quiet) {
		fprintf(stderr,"Warning: quiet mode specified without logging, you will lose your scan results\n");
	}
		
	if (o.debug > 0) {
		fprintf(stderr,"%d communities:", community_count);
		for (i=0; i < community_count; i++)
			fprintf(stderr," %s", community[i]);
		fprintf(stderr,"\n");
	}
}

/* TODO: hardcode this request */
inline int build_snmp_req(char* buf, int buf_size, char* community)
{
	int i;
	static int id;
	char snmp1_object[] = "\x30\x0e\x30\x0c\x06\x08\x2b\x06\x01\x02\x01\x01\x01\x0\x05\x00";
	char snmp2_object[] =  "\x30\x0d\x30\x0b\x06\x07\x2b\x06\x01\x02\x01\x01\x01\x05\x00";
	char *object=NULL;
	int object_size;
	int com_len;

	if (o.version == 1)
	{
		object = (char *)snmp1_object;
		object_size = sizeof(snmp1_object);
	}
	else
	{
		object = (char *)snmp2_object;
		object_size = sizeof(snmp2_object);
	}

	com_len = strlen(community);

	if (21 + com_len + strlen(object) > buf_size-1) {  // I don't know how to count.
		fprintf(stderr, "Possible boundaries overflow in build_snmp_req()?\nCommunity: %s\nObject: %s\n", community, object);
		exit(EXIT_FAILURE);
	}

	srandom(time(NULL) + getpid() + id);
	if (--id > 0x7ffe) id = rand() % 32700;

	memset(buf, 0, buf_size);

	buf[0] = 0x30;
	buf[1] = 19 + com_len + object_size-1;
	
	buf[2] = 0x02;
	buf[3] = 0x01;
	if (o.version == 1)
		buf[4] = 0x00;  // ver 1
	else
		buf[4] = 0x01;  // ver 2c
	
	// Community string
	buf[5] = 0x04;
	buf[6] = com_len;
	
	strcpy((buf + 7), community);
	i = 7 + com_len;
	
	// PDU type (get or bulk get?)
	if (o.version == 1)
		buf[i++] = 0xa0;
	else
		buf[i++] = 0xa5;

	buf[i++] = 12 + object_size-1;
	
	// Request ID
	buf[i++] = 0x02; 
	buf[i++] = 0x04;
		buf[i++] = (char)((id >> 24) & 0xff);
		buf[i++] = (char)((id >> 16) & 0xff);
		buf[i++] = (char)((id >> 8) & 0xff);
		buf[i++] = (char)((id >> 0) & 0xff);

	// Error status: no error
	buf[i++] = 0x02;
	buf[i++] = 0x01;
	buf[i++] = 0x00;
	
	// Error index
	buf[i++] = 0x02;
	buf[i++] = 0x01;
	if (o.version == 1)
		buf[i++] = 0x00;
	else
		buf[i++] = 0x0a;

	// Object ID
	memcpy((char*)&buf[i], object, object_size-1);
	i = i + object_size-1;

	return (i);
}

void _logf(char* fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	if (o.log) vfprintf(o.log_fd, fmt, args);
	if (!o.quiet) vprintf(fmt, args);
	va_end(args);
}

int parse_asn_length(u_char* buf, int buf_size, int* i)
{
	int len;
	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}
	
	if (buf[*i] < 0x81) {
		len = buf[*i];
		*i += 1;
	} else if (buf[*i] == 0x81) {
		*i += 1;
		if ((*i)+1 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		len = buf[*i];
		*i += 1;
	}
	else if (buf[*i] == 0x82) {
		*i += 1;
		if ((*i)+2 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		len = (buf[*i] << 8) + buf[(*i)+1];
		*i += 2;
	}
	else if (buf[*i] == 0x83) {
		*i += 1;
		if ((*i)+3 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		len = (buf[*i] << 16) + (buf[(*i)+1] << 8) + buf[(*i)+2];
		*i += 3;
	}
	else if (buf[*i] == 0x84) {
		*i += 1;
		if ((*i)+4 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		len = (buf[*i] << 24) + (buf[(*i)+1] << 16) + (buf[(*i)+2] << 8) + buf[(*i)+3];
		*i += 4;
	}
	else {
		_logf("Unable to decode SNMP packet: wrong length\n");
		return -1;
	}

	if ((*i)+len > buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	return len;
}

int skip_asn_length(u_char* buf, int buf_size, int* i)
{
	int ret;
	
	if ((ret = parse_asn_length(buf, buf_size, i)) > 0)
		*i += ret;

	return ret;
}

int parse_asn_integer(u_char* buf, int buf_size, int* i)
{
	int ret;

	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[*i] == 0x81) {
		*i += 1;
		if (*i >= buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
	}

	if (buf[*i] == 0x01) {
		if ((*i)+2 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		ret = (int)buf[(*i)+1];
		*i += 2;
	} else if (buf[*i] == 0x02) {
		if ((*i)+3 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		ret = 	((int)buf[(*i)+1] << 8) +
				(int)buf[(*i)+2];
		*i += 3;
	} else if (buf[*i] == 0x04) {
		if ((*i)+5 > buf_size) {
			_logf("Unable to decode SNMP packet: buffer overflow\n");
			return -1;
		}
		ret = 	((int)buf[(*i)+1] << 24) +
				((int)buf[(*i)+2] << 16) +
				((int)buf[(*i)+3] << 8) +
				(int)buf[(*i)+4];
		*i += 5;
	}
	else {
		_logf("Unable to decode SNMP packet: unrecognized integer length\n");
		return -1;
	}

	return ret;
}

int print_asn_string(u_char* buf, int buf_size, int* i)
{
	int ret;
	int string_end;

	if ((ret = parse_asn_length(buf, buf_size, i)) == -1)
		return -1;
	else
		string_end = *i + ret;

	for (;*i < string_end; *i += 1) {
		if (buf[*i] < 0x20 || buf[*i] > 0x80)
			_logf(" ");
		else
			_logf("%c", buf[*i]);
	}

	return 0;
}

int parse_snmp_header(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x30) {
		_logf("Unable to decode SNMP packet: wrong header\n");
		return -1;
	}

	if (parse_asn_length(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

int parse_snmp_version(u_char* buf, int buf_size, int* i)
{
	int ret;

	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x02) {
		_logf("Unable to decode SNMP packet: snmp version invalid\n");
		return -1;
	}

	if ((ret = parse_asn_integer(buf, buf_size, i)) == -1)
		return -1;
	else if (ret != (o.version == 1 ? 0 : 1)) {
		_logf("Unable to decode SNMP packet: snmp version invalid\n");
		return -1;
	}

	return 0;
}

int parse_snmp_community(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x04) {
		_logf("Unable to decode SNMP packet: community name not found\n");
		return -1;
	}

	_logf("[");
	if (print_asn_string(buf, buf_size, i) == -1)
		return -1;
	_logf("] ");

	return 0;
}

int parse_snmp_pdu(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0xa2) {
		_logf("Unable to decode SNMP packet: PDU type not RESPONSE (0xa2)\n");
		return -1;
	}
	
	if (parse_asn_length(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

int parse_snmp_requestid(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x02) {
		_logf("Unable to decode SNMP packet: request id invalid\n");
		return -1;
	}
	if (parse_asn_integer(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

int parse_snmp_errorcode(u_char* buf, int buf_size, int* i)
{
	int ret;

	if (*i >= buf_size) {
		_logf("(error) Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x02) {
		_logf("(error) Unable to decode SNMP packet: error code invalid\n");
		return -1;
	}
	if ((ret = parse_asn_integer(buf, buf_size, i)) < 0)
		return -1;
	if (ret != 0) {
		if (ret < 0 || ret > 18) {
			_logf("(error) Unable to decode SNMP packet: error code invalid\n");
			return -1;
		}			
		_logf("(error) Host responded with error %s\n", snmp_errors[ret]);
		return -1;
	}

	return 0;
}

int parse_snmp_errorindex(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("(error) Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x02) {
		_logf("(error) Unable to decode SNMP packet: error index invalid\n");
		return -1;
	}
	if (parse_asn_integer(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

int parse_snmp_objheader(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("(error) Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x30) {
		_logf("(error) Unable to decode SNMP packet: invalid object header\n");
		return -1;
	}
	if (parse_asn_length(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

int parse_snmp_objheader6(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("(error) Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x06) {
		_logf("(error) Unable to decode SNMP packet: invalid object header\n");
		return -1;
	}
	if (skip_asn_length(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

int parse_snmp_value(u_char* buf, int buf_size, int* i)
{
	if (*i >= buf_size) {
		_logf("(error) Unable to decode SNMP packet: buffer overflow\n");
		return -1;
	}

	if (buf[(*i)++] != 0x04) {
		_logf("(error) Unable to decode SNMP packet: invalid value\n");
		return -1;
	}
	if (print_asn_string(buf, buf_size, i) < 0)
		return -1;

	return 0;
}

void parse_snmp_response(u_char* buf, int buf_size)
{
	int i;

	i = 0;

	if (parse_snmp_header(buf, buf_size, &i) == -1) return;
	if (parse_snmp_version(buf, buf_size, &i) == -1) return;
	if (parse_snmp_community(buf, buf_size, &i) == -1) return;
	if (parse_snmp_pdu(buf, buf_size, &i) == -1) return;
	if (parse_snmp_requestid(buf, buf_size, &i) == -1) return;
	if (parse_snmp_errorcode(buf, buf_size, &i) == -1) return;
	if (parse_snmp_errorindex(buf, buf_size, &i) == -1) return;
		
	if (i+3 <= buf_size && buf[i] == 0x00 && buf[i+1] == 0x30 && buf[i+2] == 0x20)	// Bug in an HP JetDirect
		i += 3;
	
	if (parse_snmp_objheader(buf, buf_size, &i) == -1) return;
	if (parse_snmp_objheader(buf, buf_size, &i) == -1) return;		// yes, this should be called twice
	if (parse_snmp_objheader6(buf, buf_size, &i) == -1) return;
	if (parse_snmp_value(buf, buf_size, &i) == -1) return;

    _logf("\n");
}

/* Subtract the `struct timeval' values X and Y,
 * storing the result in RESULT.
 * Return 1 if the difference is negative, otherwise 0.
 */

inline int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y)
{
	int nsec;

	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	/* Compute the time remaining to wait. tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;
	
	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}

void receive_snmp(int sock, long wait)
{
	struct timeval tv_now, tv_until, tv_wait;
	int remote_addr_len;
	char buf[2000];
	int ret;
	fd_set fds;
	struct sockaddr_in remote_addr;
	int errcount=0;

	memset(&remote_addr, 0, sizeof(struct sockaddr_in));

	gettimeofday(&tv_now, NULL);
	tv_until.tv_sec = tv_now.tv_sec;
	tv_until.tv_usec = tv_now.tv_usec + wait * 1000;
	if (tv_until.tv_usec >= 1000000) {
		tv_until.tv_sec += tv_until.tv_usec / 1000000;
		tv_until.tv_usec = tv_until.tv_usec % 1000000;
	}

	tv_wait.tv_sec = wait / 1000;
	tv_wait.tv_usec = wait % 1000 * 1000;
	
	do {
		/* Put the socket into the fd set */
		FD_ZERO(&fds);
		FD_SET(sock, &fds);
	
		if ((ret = select(sock+1, &fds, NULL, NULL, &tv_wait)) == -1) {
			perror("select()");
			exit(EXIT_FAILURE);
		} else if (ret > 0) {
			remote_addr_len = sizeof(remote_addr);
	
			ret = recvfrom(sock, &buf, sizeof(buf), 0, (struct sockaddr*)&remote_addr, (socklen_t *)&remote_addr_len);
			if (ret < 0) {
				perror("recvfrom()");
				++errcount;
				if (errcount >= MAX_ERROR) { fprintf(stderr, "Too many errors (recvfrom); exiting.\n"); fflush(o.log_fd); fclose(o.log_fd); exit(EXIT_FAILURE); }
			}
			_logf("%s ", inet_ntoa(remote_addr.sin_addr));
			parse_snmp_response((u_char*)&buf, ret);
			if (o.log) fflush(o.log_fd);
		}

		gettimeofday(&tv_now, NULL);
	} while (timeval_subtract(&tv_wait, &tv_until, &tv_now) == 0);
}

int main(int argc, char* argv[])
{
	struct sockaddr_in local_addr;
	struct sockaddr_in remote_addr;
	int sock;
	int ret;
	int c, i, a;
	char sendbuf[1500];
	int sendbuf_size;
	unsigned int errcount;
	int sleepval=USLEEP_TIME;

	printf("+-+-+-+-+-+-+--+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n");
	printf("| 161plus.c by xt; your friendly ultra-fast snmp scanner |\n");
	printf("+-+-+-+-+-+-+--+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+\n\n");

	srandom(time(NULL) + getpid());

	init_options(argc, argv);

	/* socket creation */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		perror("socket()");
		exit(EXIT_FAILURE);
	}

	local_addr.sin_family = AF_INET;
	local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	local_addr.sin_port = htons(0);


	ret = bind(sock, (struct sockaddr *)&local_addr, sizeof(local_addr));
	if (ret < 0) {
		perror("bind()");
		exit(EXIT_FAILURE);
	}
	
	/* remote address */
	remote_addr.sin_family = AF_INET;
	remote_addr.sin_port = htons(161);

	printf("hosts: %lu, communities: %d\n", host_count, community_count);
	printf("select pkt delay: %d, select pkt interval: %d, usleep interval: %d\n", o.packets, o.packets, o.usl);

	signal(SIGINT, &sig_catch);
	signal(SIGHUP, &sig_catch);
	signal(SIGTERM, &sig_catch);
	
	for (c=0; c < community_count; c++) {
		if (o.debug > 0) fprintf(stderr,"Scanning community \'%s\'\n", community[c]);

		sendbuf_size = build_snmp_req((char*)&sendbuf, sizeof(sendbuf), community[c]);

		for (errcount=0,i=0,a=0; i < host_count; i++,a++) {
			remote_addr.sin_addr.s_addr = hosts[i];
			if (o.debug > 1) printf("Sending to ip %s\n", inet_ntoa(*(struct in_addr*)&remote_addr.sin_addr.s_addr));

			ret = sendto(sock, &sendbuf, sendbuf_size, 0, (struct sockaddr*)&remote_addr, sizeof(remote_addr));
			if (ret < 0) {
				perror("sendto()");
				if (++errcount > MAX_ERROR) { fprintf(stderr, "Too many errors (sendto); exiting.\n"); fflush(o.log_fd); fclose(o.log_fd); exit(EXIT_FAILURE); }
			}

			// let's see if this speeds up the scan a bit
			if (!(a % o.packets)) {
				receive_snmp(sock, o.wait);
			}
			else if (!(a % o.usl)) { usleep(sleepval); }
		}
	}
	
	receive_snmp(sock, o.wait);
	if (o.debug > 0) fprintf(stderr, "All packets sent, waiting for responses.\n");

	/* wait for 5 seconds */
	receive_snmp(sock, 5000);

	if (o.debug > 0) fprintf(stderr,"done.\n");

	if (o.log) fclose(o.log_fd);

	free(hosts);

	return 0;
}

/*  This function courtesey of Ben Pfaff (http://www.benpfaff.org/) */
void array_shuffle(u_int32_t *array, size_t n)
{
	if (n > 1)
	{
		size_t i;

		for (i = 0; i < n - 1; i++)
		{
			size_t j = i + random() / (RAND_MAX / (n - i) + 1);
		 	u_int32_t t = array[j];
			array[j] = array[i];
			array[i] = t;
	  	}
	}
}

void sig_catch(int sigtype)
{
	switch (sigtype)
	{
		case SIGINT:
			fflush(o.log_fd);
			fclose(o.log_fd);
			fprintf(stderr, "\n ... Exiting cleanly at user's request. Flushed and wrote log file.\n");
			exit(EXIT_SUCCESS);
			break;
		case SIGTERM:
		case SIGHUP:
			fflush(o.log_fd);
			fclose(o.log_fd);
			fprintf(stderr, "\n ... Caught SIGTERM or SIGHUP; exiting cleanly. Flushed and wrote log file.\n");
			break;
		default:
			break;
	}
}
